# 
# ZCVS Mixin Base Class
# 
# Steve Spicklemire
# Silicon Prairie Ventures Inc. (SPVI)
# 
# Disclaimer
# 
#   
#   THIS SOFTWARE IS PROVIDED BY SPVI ``AS IS'' AND ANY EXPRESSED OR
#   IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#   DISCLAIMED.  IN NO EVENT SHALL SPVI OR ITS CONTRIBUTORS BE LIABLE FOR
#   ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
#   GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
#   IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
#   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#   
# 
# 

__doc__ = "ZCVS Mixin Base Class"
__version__ = "0.1.4"

from ZCVSExIm import doExport
from ShellCommandProcessor import ShellCommandProcessor
import StringIO
import traceback

try:

    from OFS import XMLExportImport
    customImporters={
        XMLExportImport.magic: XMLExportImport.importXML,
    }
except:
    customImporters = None

try:
    import Globals, OFS, AccessControl, zLOG
    from Acquisition import Implicit
    from Globals import Persistent, HTMLFile, MessageDialog, ImageFile
    from DocumentTemplate import html_quote
    
    def do_break():
        import pdb
        pdb.set_trace()

    class Dummy (Persistent,
                 Implicit,
                 AccessControl.Role.RoleManager,
                 ShellCommandProcessor):
        pass

except:

    class foo:
        pass

    class HTMLFile:

        def __init__(self, *args):
            pass

    class ImageFile:
        def __init__(self, *args):
            pass

    def LOG( subsystem, severity, short ):
        if severity > -500:
            print subsystem, severity, short

    zLOG = foo()
    zLOG.LOG = LOG
    zLOG.BLATHER = 0
    zLOG.INFO = 1

    class Dummy(ShellCommandProcessor):
        """ for testing outside Zope... """

        def objectIds(self):
            return ['hehe','nuther']

        hehe = 'foo'
        nuther = 'blah'

        def _getOb(self, id):
            return None


import os
import string
import time
import tempfile

statDict = {
    'Up-to-date':'OK',
    'Locally Modified':'LM',
    'Locally Added':'LA',
    'Locally Removed':'LR',
    'Needs Checkout':'NC',
    'Needs Patch':'NP',
    'Unknown':'UN',
    }

class TimeTracker:

    """
    Keep track of times... we have:

    fsmtime  (mtime from filesystem...) and
    bmtime  (bobbase modification time for 'real' modification)

    When an object is pulled from cvs it gets a bobobase
    modification time of 'now'. CVS also has a concept of
    modification time. Assuming nobody monkeys with files
    on the filesystem other than cvs.. we can 'fake' real
    modification times by deliberately keeping track of
    fsmtime and bmtime.

    """

    fsmtime = 0.0
    bmtime = 0.0

    def ZopeChanged(self, curr_z_bmtime):
        """ put in 5 seconds leeway.. """
        return curr_z_bmtime > self.bmtime + 5

    def FSChanged(self, curr_fs_mtime):
        """ put in 5 seconds leeway.. """
        return curr_fs_mtime > self.fsmtime + 5
            
    def ZopeChangedTime(self, curr_z_bmtime):

        if self.ZopeChanged():
            return curr_z_bmtime
        else:
            return self.fsmtime

    def FSChangedTime(self, curr_fs_mtime):
        if self.FSChanged(curr_fs_mtime):
            return curr_fs_mtime
        else:
            return self.fsmtime


DEFAULTPASSFILE='/var/zope/.cvspass'

class ZCVSMixin(Dummy):

    """ZCVS Mixin"""

    meta_type = "ZCVS Mixin Base Class"
    fsroot = '/tmp'
    fsrepos = ''
    addfsdir = 1
    isInZClass = 0  # special tricks for ZClasses...
    getFSRootFromParent = 1  # try to get FSRoot from parent class
    default_gargs = '-z6'
    trackerDict = None
    passfile = DEFAULTPASSFILE
    usepassfile = 1
    badTypes = ('ZCVS Folder', 'Folder', 'CVS Folder') # we don't do these (yet?)
    excludeTypes = []
    _v_debug = 0

    __ac_permissions__ = (
        
        ('Manage CVS Controls',
         ['manage_checkIn', 'manage_checkOut', 'cvsAdminForm','cvsStatusForm',
          'manage_addfsdirCVS', 'manage_statusCVS', 'manage_addCVS',
          'manage_editCVS', 'manage_cvsAdmin','manage_objFromFS',
          'manage_objToFS'], ['Manager']),
	)

    manage_options = (
	{'label':'CVS Status', 'action':'cvsStatusForm'},
	{'label':'CVS Admin', 'action':'cvsAdminForm'},
	)

    cvsAdminForm = HTMLFile('cvsAdminForm', globals())
    cvsStatusForm = HTMLFile('cvsStatusForm', globals())

    same_icon = ImageFile('same.gif', globals())

    left_blue_icon = ImageFile('left_blue.gif', globals())
    left_green_icon = ImageFile('left_green.gif', globals())
    left_red_icon = ImageFile('left_red.gif', globals())

    right_blue_icon = ImageFile('right_blue.gif', globals())
    right_green_icon = ImageFile('right_green.gif', globals())
    right_red_icon = ImageFile('right_red.gif', globals())

    def excludedTypes(self):
        return list(self.badTypes) + self.excludeTypes

    def _containerObjectIds(self):
        return self.objectIds()

    def _containerHasID(self, id):
        return id in self.objectIds()

    def _container_p_jar(self):
        return self._p_jar

    def _containerDelObject(self, id):
        self._delObject(id)

    def _containerSetObject(self, id, value):
        self._setObject(id, value)

    def _containerGetattr(self, id):
        return getattr(self, id)

    def checkTracker(self):
        if self.trackerDict is None:
            self.trackerDict = {}

    def reloadAdmin(self):
        """ dynamic reload of admin forms... for development.. """
        self.cvsAdminForm = HTMLFile('cvsAdminForm', globals())
        self.cvsStatusForm = HTMLFile('cvsStatusForm', globals())
        return "OK!"

    def getFSRoot(self, PARENTS=None):

        result = None
        
        if self.getFSRootFromParent:
            if PARENTS:
                for index in range(len(PARENTS)):
                    try:
                        result = PARENTS[index].__dict__.get('CVS',None).fsroot  # get the highest possible 'root'
                    except AttributeError:
                        pass
            else:
                #do_break()
                try:
                    result = self.aq_parent.aq_parent.CVS.getFSRoot()
                except AttributeError:
                    pass

            if result is not None:
                return result
        
        if hasattr(self, 'fsdir'): # backward compatibility.....
            return os.path.split(self.fsdir)[0]
        else:
            return self.fsroot

    def getFSRepos(self):
        if hasattr(self, 'fsdir'):
            return '/' + os.path.split(self.fsdir)[1]
        else:
            return self.fsrepos

    def getMetaType(self, obj):
        """ get a meta type.... call if necessary """
        result = ''
        
        if hasattr(obj, 'meta_type'):
            mt = obj.meta_type
            if hasattr(mt, 'im_func'):
                mt = mt()
            result = mt

        return result
    
    def getNewFSRepos(self, PARENTS=None):
        """ try to determine the new repository name looking into Parents... """

        base = ''
        isZClass = 0

        mt = self.getMetaType(PARENTS[1])
        
        if mt == 'Z Class':
            #
            # Ack... special case... fix up for ZClass
            #
            foundIndex = -1
            isZClass = 1
            
            for obj in PARENTS:
                if obj.__class__.__name__ == 'Product':
                    break
                elif obj.__class__.__name__ == 'ZClass':

                    if base:
                        base = os.sep + base
                        
                    base = obj.id + base
        else:
            base = PARENTS[1].id

        for index in range(len(PARENTS)):
            try:
                base = PARENTS[index].__dict__.get('CVS',None).fsrepos + os.sep + base
                break
            except AttributeError:
                pass
                
        return base, isZClass
    
    def getFSDir(self):
        if hasattr(self, 'fsdir'):
            return self.fsdir
        else:
            return self.getFSRoot() + self.getFSRepos()

    def checkList(self, obj_id):
        """
        check for list type... convert if necessary.
        """

        if type(obj_id) == type(''):
            obj_id = string.split(obj_id)
        elif obj_id is None:
            obj_id = []

        return obj_id

    def idsToFileNames(self, statusDict, relNames=None):
        fnames = []
        for obj_id in statusDict.keys():
            if statusDict.has_key(obj_id) and statusDict[obj_id].has_key('ext'):
                ext = statusDict[obj_id]['ext']
            else:
                ext = 'xml'
                
            if relNames:
                fnames.append("%s.%s" % (obj_id,ext))
            else:
                fnames.append(os.path.join(self.getFSDir(), ("%s.%s" % (obj_id, ext))))

        return fnames

    def idToFileName(self, obj_id, ext, relName=None):

        if relName:
            return "%s.%s" % (obj_id, ext)
        else:
            return os.path.join(self.getFSDir(), "%s.%s" % (obj_id, ext))

    def parseStatusList(self, resultStd, secondary=None):
        """ parse stdout from 'cvs status'... use any clues
            we can find to figure out what's what. Results
            are returned in a dict of dicts that has object_ids
            as keys and all the info gleaned from cvs in
            a sub-dictionary tied to that key.
        """

        results = {}
        currObject = {}
        self.checkTracker()

        td = self.trackerDict

        # do_break()
        
        if secondary:
            fnames = []  # we don't care about the file system.. this is a secondary
            zobjs = []
        else:
            fnames = os.listdir(os.curdir)
            zobjs = self._containerObjectIds()

            if 'CVS' not in fnames:
                #
                # we are not in a cvs directory.. punt...
                #
                return results
        

        if self._v_debug:
            zLOG.LOG('ZCVSMixin', zLOG.INFO , 'zobjs..\n' + `zobjs`)
            zLOG.LOG('ZCVSMixin', zLOG.INFO , 'fnames..\n' + `fnames`)
            zLOG.LOG('ZCVSMixin', zLOG.INFO , 'result from cvsStatus...\n' + resultStd)
        
        for line in string.split(resultStd,'\n'):
            if string.find(line, '==========') != -1:
                if currObject.has_key('zname'):
                    results[currObject['zname']] = currObject
                    currObject = {}

            elif string.find(line, 'Status:') != -1:
                status = string.strip(string.split(line, 'Status:')[1])
                currObject['status'] = statDict.get(status, status)
                fname = string.strip(string.split(string.split(line, 'File:')[1],'Status:')[0])
                if string.find(fname,'no file') == -1:
                    currObject['fname'] = fname
                    currObject['zname'] = string.join(string.split(fname,'.')[:-1],'.')
                    currObject['ext'] = string.split(fname,'.')[-1]
                
            elif string.find(line, 'Working revision') != -1:
                currObject['wRev'] = string.strip(string.split(line,':')[1])
            elif string.find(line, 'Repository revision') != -1:
                currObject['rRev'] = rRev = string.strip(string.split(string.split(line,':')[1])[0])
                if rRev == 'No':
                    currObject['rRev'] = rRev = string.strip(string.split(line,':')[1])
                elif not currObject.has_key('fname'):
                    fname = string.strip(string.join(string.split(string.split(line,':')[1])[1:]))
                    fname = string.join(string.split(os.path.split(fname)[-1],',')[:-1],',')
                    currObject['fname'] = fname
                    currObject['zname'] = string.join(string.split(fname,'.')[:-1],'.')
                    currObject['ext'] = string.split(fname,'.')[-1]

            elif string.find(line, 'Sticky Tag') != -1:
                currObject['sTag'] = string.strip(string.split(line,':')[1])
            elif string.find(line, 'Sticky Date') != -1:
                currObject['sDate'] = string.strip(string.split(line,':')[1])
            elif string.find(line, 'Sticky Options') != -1:
                currObject['sOptions'] = string.strip(string.split(line,':')[1])

        if currObject.has_key('zname'):
            results[currObject['zname']] = currObject

        for fname in fnames:
            theSplit = string.split(fname,'.')
            zobj_id = string.join(theSplit[:-1],'.')
            ext = theSplit[-1]
            if self._v_debug:
                zLOG.LOG('ZCVSMixin', zLOG.INFO, 'looking at %s, %s:%s:%s' % (fname, `theSplit`, zobj_id, ext))
            
            if not results.has_key(zobj_id) and (len(theSplit)>1) and (ext in ('xml', 'zexp')):
                currObject = {}
                currObject['zname'] = zobj_id
                currObject['status'] = 'OR'
                currObject['fname'] = fname
                currObject['ext'] = ext
                currObject['format_opt'] = ext
                results[zobj_id] = currObject

        for id in zobjs:
            if not results.has_key(id):
                results[id] = {'zname':id}

        for k in results.keys():

            tracker = td.get( string.strip(k), TimeTracker() )
            results[k]['tracker'] = tracker
            
            if k in zobjs:

                # get the object and see if it's changed....
                
                results[k]['zobj'] = theObj = self._containerGetattr(k)
                if hasattr(theObj, 'meta_type'):
                    results[k]['meta_type'] = theObj.meta_type

                if hasattr(theObj, 'bobobase_modification_time'):
                    btime = theObj.bobobase_modification_time

                    if hasattr(btime, 'im_func'):
                        btime = btime()

                    results[k]['btime'] = btime = btime.timeTime()
                    results[k]['znewerf'] = tracker.ZopeChanged( btime )

                # check for extension options...
                
                if not results[k].has_key('ext'):
                    if (hasattr(theObj, 'isPrincipiaFolderish') and theObj.isPrincipiaFolderish) \
                       or theObj.meta_type == 'Image':
                        
                        results[k]['format_opt'] = 'zexp'
                    else:
                        results[k]['format_opt'] = 'xml'
                else:
                    if results[k].has_key('fname') and os.path.exists(results[k]['fname']):
                        results[k]['mtime'] = mtime = os.path.getmtime(results[k]['fname'])

                        if tracker.FSChanged(mtime):
                            results[k]['fnewerz'] = 1
            else:
                results[k]['zobj'] = None

            if results[k].has_key('status'):
                if (results[k]['status'] in ('LM', 'LA')):
                    results[k]['fnewerc'] = 1
                if (results[k]['status'] in ('NU', 'NP')):
                    results[k]['cnewerf'] = 1
            else:
                results[k]['status'] = 'UN'


            for k in results.keys():
                if results[k].has_key('meta_type') and (results[k]['meta_type'] in self.excludedTypes()):
                    del results[k]
                
        return results

    def parseUpdateList(self, resultStd):
        """ parse stdout from cvs update... """
        
        results = {}
        lines = string.split(resultStd, '\n')
        for line in lines:
            tokens = string.split(line)
            if len(tokens)>1:
                status = tokens[0]
                fname = tokens[1]
                results[fname] = status

        return results

    def cvsStatusList(self):
        """
        Return a dictionary of 'status' settings for objects in CVS compared to
        files on the filesystem. Most other functions take dictionaries like
        this to decide what to do....
        """

        oldDir = os.getcwd()  # move to CVS directory....
        try:
            os.chdir(self.getFSDir())
        except:
            return (0, {}, 'No such dir')
                 
        resultVal, resultStd, resultErr = self.processStreams( "cvs %s -n update -l" % self.default_gargs )  # check the update

        updateList = self.parseUpdateList( resultStd )  # get the update list.. don't really update...
        
        resultVal, resultStd, resultErr = self.processStreams( "cvs %s status -l" % self.default_gargs )  # check the status

        result = self.parseStatusList( resultStd)

        statAgainList = []
        for fname in updateList.keys():
            obj_id = string.join(string.split(fname,'.')[:-1],'.')
            if not result.has_key(obj_id):
                statAgainList.append( fname )

        if statAgainList:
            resultVal, resultStd, resultErr = self.processStreams(
                'cvs %s status -l %s ' % (self.default_gargs, quoteArgs( statAgainList )))

            secondaryResult = self.parseStatusList( resultStd, secondary=1)

            for k in secondaryResult.keys():
                result[k] = secondaryResult[k]
                
        os.chdir(oldDir)

        return resultVal, result, resultErr

    def manage_editCVS(self, fsdir=None, fsroot=None, fsrepos=None, addfsdir=0, usepassfile=0,
                       active=0, passfile=DEFAULTPASSFILE, default_gargs='', excludeTypes=None,
                       getFSRootFromParent=0, REQUEST=None):
        """
        Check the properties of the cvs object...
        """

        if hasattr(self, 'fsdir'):
            self.fsdir = fsdir  # backward compatable for old instances...
        else:
            self.fsroot = fsroot
            self.fsrepos = fsrepos
            
        self.addfsdir = addfsdir
        self.usepassfile = usepassfile
        self.passfile=passfile
        self.default_gargs = default_gargs
        self.active = active
        self.getFSRootFromParent = getFSRootFromParent

        if excludeTypes:
            self.excludeTypes = excludeTypes

        return MessageDialog(title='OK',
			     message='CVS Folder Params Edited.',
			     action='cvsAdminForm')

    def manage_checkOut(self, obj_ids=None, g_args='', cvs_args='', force_update=0, REQUEST=None):
        """
        check out the latest version...

        1) move the latest changes in zope to the filesystem
           (use the btime date/tracker to decide if changes have been made)

        2) cvs update.. with args...

        3) load updated files into Zope..
        
        """

        oldDir = os.getcwd()  # move to CVS directory....
        os.chdir(self.getFSDir())
        g_args = g_args + ' ' + self.default_gargs
        force_specified_files = 1  # update the files no matter what... 

        rv, statDict, ev = self.cvsStatusList()

        obj_id_list = self.checkList(obj_ids)

        if len(obj_id_list) == 0:
            force_specified_files = 0
            obj_id_list = statDict.keys()
            
        saveDict = {}
        
        for obj_id in obj_id_list:
            if force_specified_files or statDict[obj_id].has_key('znewerf') and statDict[obj_id]['znewerf']:
                saveDict[obj_id] = statDict[obj_id]
                
        self.objectsToFiles(saveDict, force_time=force_specified_files)      # save to FS

        rv, statDict, ev = self.cvsStatusList()

        updateList = []
        updateDict = {}

        for obj_id in obj_id_list:
            curStat = statDict[obj_id].get('status','na')
            if force_specified_files or curStat == 'NP' or curStat == 'NC':
                updateList.append( statDict[obj_id]['fname'] )
                updateDict[obj_id] = statDict[obj_id]
                        
        command = 'cvs %s update -l %s %s ' % (g_args, cvs_args, quoteArgs(updateList))
        
        iresult, resultStd, resultErr = self.processStreams( command )  # do the update

        result = self.formatResult( command, iresult, resultStd, resultErr) + '<br>' 
        result = result + self.filesToObjects( updateDict )
        
        result = "<table><tr><td><pre>%s</pre></td></tr></table>" % result

        os.chdir(oldDir)

        return MessageDialog(title='Check Out Returned',
			     message='command: %s <br> Update returned: "%s"' % (srj(command), result),
			     action='cvsStatusForm')


    def manage_objFromFS(self, obj_ids=None, REQUEST=None):
        """
        move objects from files to Zodb

        """

        oldDir = os.getcwd()  # move to CVS directory....
        os.chdir(self.getFSDir())

        rv, statDict, ev = self.cvsStatusList()

        obj_id_list = self.checkList(obj_ids)

        if len(obj_id_list) == 0:
            obj_id_list = statDict.keys()
            
        updateDict = {}

        for obj_id in obj_id_list:
            if statDict.has_key(obj_id) and statDict[obj_id].has_key('fname'):
                updateDict[obj_id] = statDict[obj_id]
        
        result = self.filesToObjects( updateDict )
        
        result = "<table><tr><td><pre>%s</pre></td></tr></table>" % result

        os.chdir(oldDir)

        return MessageDialog(title='Check Out Returned',
			     message='Update returned: "%s"' %  result,
			     action='cvsStatusForm')


    def manage_checkIn(self, obj_ids=None, message="No Message", g_args='', cvs_args='', REQUEST=None):
        """
        check in the current version(s)...
        """

        oldDir = os.getcwd()  # move to CVS directory....
        os.chdir(self.getFSDir())
        g_args = g_args + ' ' + self.default_gargs
        force_specified_files = 1

        rv, statDict, ev = self.cvsStatusList()

        obj_id_list = self.checkList(obj_ids)

        if len(obj_id_list) == 0:
            force_specified_files = 0
            obj_id_list = statDict.keys()
            
        saveDict = {}
        
        for obj_id in obj_id_list:
            ld = statDict[obj_id]
            if ld.has_key('meta_type') and (ld['meta_type'] not in self.excludedTypes()):
                if force_specified_files or (ld.has_key('znewerf') and ld['znewerf']):
                    saveDict[obj_id] = ld

        self.objectsToFiles(saveDict, force_time=force_specified_files) # save to FS
        
        fnames = self.idsToFileNames( saveDict, relNames=1 )       # convert to fnames

        command = 'cvs %s commit -l -m %s %s %s' % (g_args, "'" + message + "'", cvs_args, quoteArgs(fnames))

        result = self.processCommand( command )

        rv, statDict, ev = self.cvsStatusList()

        restoreDict = {}
        
        for obj_id in obj_id_list:
            ld = statDict[obj_id]
            if ld.has_key('meta_type') and (ld['meta_type'] not in self.excludedTypes()):
                if ld.has_key('fnewerz') and ld['fnewerz']:
                    restoreDict[obj_id] = ld

        result = result + self.filesToObjects( restoreDict )
        
        result = "<table><tr><td><pre>%s</pre></td></tr></table>" % result

        os.chdir(oldDir)

        return MessageDialog(title='Commit Returned',
			     message='command: %s <br> Commit returned... "%s"' % (srj(command), result),
			     action='cvsStatusForm')

    def manage_objToFS(self, obj_ids=None, REQUEST=None):
        """
        check in the current version(s)...
        """
           
        oldDir = os.getcwd()  # move to CVS directory....
        os.chdir(self.getFSDir())

        rv, statDict, ev = self.cvsStatusList()

        obj_id_list = self.checkList(obj_ids)

        if len(obj_id_list) == 0:
            obj_id_list = statDict.keys()
            
        saveDict = {}
        
        for obj_id in obj_id_list:
            ld = statDict[obj_id]
            if ld.has_key('meta_type') and (ld['meta_type'] not in self.excludedTypes()):
                saveDict[obj_id] = ld

        self.objectsToFiles(saveDict)                  # save to FS
        os.chdir(oldDir)
        
        return MessageDialog(title='Object Converted to Files',
			     message='Objects saved to files... %s ' % string.join(saveDict.keys(),'<br>'),
			     action='cvsStatusForm')

    def manage_addfsdirCVS(self, g_args=''):
        """
        add dir to CVS
        """

        oldDir = os.getcwd()  # move to CVS directory....

        if not os.access(self.getFSDir(), os.F_OK):
            os.mkdir(self.getFSDir())

        os.chdir(self.getFSDir())

        # add directory

        result = '' 

        if not os.access(self.getFSDir() + '/CVS', os.F_OK):

            os.chdir(os.path.dirname(self.getFSDir()))
            bname = os.path.basename(self.getFSDir())
            command = 'cvs %s add %s' % (g_args, bname)
            result = self.processCommand( command )
            command = 'cvs %s commit -l -m "add folder %s" %s' % (g_args, bname, bname)
            result = result + self.processCommand( command )
        
            result = "<table><tr><td><pre>%s</pre></td></tr></table>" % (result)

        os.chdir(oldDir)

        return result
    
    def manage_addCVS(self, g_args='', obj_ids=None, add_ids=None, cvs_args='', REQUEST=None):
        """
        check out the latest version...
        """

        oldDir = os.getcwd()  # move to CVS directory....
        os.chdir(self.getFSDir())
        g_args = g_args + ' ' + self.default_gargs

        rv, statDict, ev = self.cvsStatusList()

        obj_id_list = self.checkList(obj_ids)
        add_id_list = self.checkList(add_ids)

        saveDict = {}

        if len(obj_id_list) == 0:
            obj_id_list = statDict.keys()

        realIds = []
        for theId in obj_id_list:
            if theId in add_ids:
                realIds.append(theId)
                
        obj_id_list = realIds
        
        for obj_id in obj_id_list:
            saveDict[obj_id] = statDict[obj_id]
            saveDict[obj_id]['ext'] = saveDict[obj_id]['format_opt'] = REQUEST.form['format;'+obj_id]

        self.objectsToFiles(saveDict)                  # save to FS
        
        if saveDict:
            #
            # need to do two different adds..... one for xml, another for zexp...
            #

            xmlDict = {}
            zexpDict = {}
            for k in saveDict.keys():
                if saveDict[k]['ext'] == 'xml':
                    xmlDict[k] = saveDict[k]
                else:
                    zexpDict[k] = saveDict[k]

            if len(xmlDict.keys())>0:
                addList = self.idsToFileNames( xmlDict, relNames=1)
                command = 'cvs add %s %s ' % (cvs_args, quoteArgs(addList))
                result1 = self.processCommand( command )  # do the update
            else:
                result1 = ''

            if len(zexpDict.keys())>0:
                addList = self.idsToFileNames( zexpDict, relNames=1)
                command = 'cvs %s add -kb %s %s ' % (g_args, cvs_args, quoteArgs(addList))
                result2 = self.processCommand( command )  # do the update
            else:
                result2 = ''


            result = "<table><tr><td><pre>%s</pre></td></tr></table>" % (result1 + result2)
        else:
            result = "No Objects to add.."

        return MessageDialog(title='Add Returned',
			     message='Add returned: "%s"' %  result,
			     action='cvsStatusForm')

        os.chdir(oldDir)

    def manage_cvsAdmin(self, cvs_admin_args='', g_args='', adminCommand='', obj_ids=None, REQUEST=None):
        """
        perform some cvs administration
        """

        if REQUEST is None:
            REQUEST = {}
            
        oldDir = os.getcwd()
        os.chdir(self.getFSDir())
        g_args = g_args + ' ' + self.default_gargs

        zLOG.LOG('ZCVSMixin', zLOG.BLATHER, 'We are in cvsAdmin')

        obj_id_list = self.checkList(obj_ids)
        rv, statDict, ev = self.cvsStatusList()

        if len(obj_id_list) == 0:
            obj_id_list = statDict.keys()

        adminDict = {}
        
        for obj_id in obj_id_list:
            adminDict[obj_id] = statDict[obj_id]

        fnames = self.idsToFileNames( adminDict, relNames=1 )       # convert to fnames

        if adminCommand == 'removeFS':
            command = 'rm ' +  quoteArgs(fnames)
        elif adminCommand == 'removeZ':
            for itemId in adminDict.keys():
                self._containerDelObject( itemId )
            result = "Deleted IDs: %s" % `adminDict.keys()`
            command = ''
        else:
            command = ('cvs -l %s %s %s' % (g_args, adminCommand, cvs_admin_args)) + quoteArgs(fnames)

        if command:
            result = self.processCommand( command )  # do the update

        result = "<table><tr><td><pre>%s</pre></td></tr></table>" % html_quote(result)
        os.chdir(oldDir)

        return MessageDialog(title='Admin Returned',
			     message='command: %s <br> Returned: "%s"' % (srj(command), result),
			     action='cvsStatusForm')


    def filesToObjects(self, obj_statDict):
        """
        import set of files from dict into this container..
        """
        failedImports = []
        failedMessages = []
        successfulImports = []
        
        self.checkTracker()  # make sure we have a trackerDict...

        #do_break()
        
        for k in obj_statDict.keys():
            if obj_statDict[k].has_key('ext'):
                ext = obj_statDict[k]['ext']
                fname = self.idToFileName(k,ext)
                try:
                    try:
                        ob=self._container_p_jar().importFile(fname, customImporters=customImporters)
                    except:
                        ob=self._container_p_jar().importFile(fname)
                        
                    id=ob.id
                    if hasattr(id, 'im_func'):
                        id=id()

                    if self._containerHasID(id):
                        self._containerDelObject(id)

                    self._containerSetObject(id, ob)

                    btime = self._containerGetattr(id).bobobase_modification_time

                    if hasattr(btime, 'im_func'):
                        btime = btime()

                    btime = btime.timeTime()

                    td = self.trackerDict
                    tracker = td.get(string.strip(id), TimeTracker())
                    tracker.bmtime = btime
                    tracker.fsmtime = os.path.getmtime(fname)
                    td[string.strip(id)] = tracker
                    self.trackerDict = td

                    successfulImports.append( fname )
                except:
                    failedImports.append(fname)
                    f = StringIO.StringIO()
                    traceback.print_exc(file=f)
                    f.seek(0)
                    failedMessages.append(f.read())

        result = "<br><b>Successful Imports:</b>\n" + string.join(successfulImports,',\n') + "<br>\n" + \
                 "<b>Failed Imports</b>\n" + string.join(failedImports,',\n') + "<br>\n" + "<br>\n" + \
                 "<b>Failed Messages</b>\n" + string.join(failedMessages,'<p><hr><p>') + "<br>\n" + "<br>\n" 
        
        return result

    def objectsToFiles(self, obj_statDict, force_time=0):
        """
        move objects from statDict from this Zope container to file system,
          return list of filenames... set modification time of
          object to be the same as the bobobase modification time
          if the object has been modified, otherwise set it to
          the old fstime. Make sure there is an extension for each file...
          otherwise it will default to zexp....
        """

        self.checkTracker()
        
        fnames = []
        td = self.trackerDict
        obj_id_list = obj_statDict.keys()

        for obj_id in obj_id_list:
            currItem = obj_statDict[obj_id]
            ob=currItem['zobj']
            
            if (currItem.has_key('ext') and currItem['ext'] == 'xml') or \
               (currItem.has_key('format_opt') and currItem['format_opt'] == 'xml'):
                fname = self.idToFileName(obj_id, 'xml')
                if ob is not None:
                    doExport(ob._p_jar, ob._p_oid, fname)
                else:
                    os.remove(fname)
            else:
                fname = self.idToFileName(obj_id, 'zexp')
                if ob is not None:
                    ob._p_jar.exportFile(ob._p_oid, fname)
                else:
                    os.remove(fname)
            
            if ob is not None:
                btime = ob.bobobase_modification_time  # keep mtime same as btime...
            else:
                btime = 0.0
            if hasattr(btime, 'im_func'):
                btime = btime().timeTime()

            tracker = td.get(obj_id, TimeTracker())

            if tracker.bmtime < btime:
                tracker.bmtime = btime  # this one has *really* changed since import
                td[string.strip(obj_id)] = tracker        # update trackerDict...
                fstime = btime
            else:
                fstime = tracker.fsmtime  # not a real change.. just import itself

            if not force_time:
                os.utime(fname, (fstime, fstime))
            
            fnames.append(fname)

        self.trackerDict = td
        return fnames

    def getTrackerDict(self):

        """
        Get the tracker dict....

        """
        self.checkTracker()
        
        return self.trackerDict

    def formatResult(self, command, resultVal, resultStd, resultErr):
        
        result = ["Command: '%s' returns %i" % (srj(command), resultVal)]
        result.append("StdOut: '%s'" % resultStd)
        result.append("StdErr: '%s'" % resultErr)

        return string.join(result,'\n')

def quoteArgs( aList ):
    return string.join(map(lambda x: "'%s'" % x, aList), ' ')

def splitResizeJoin( aString, maxSize = 70 ):

    words = string.split(aString)
    newLines = []
    currString = ''

    for word in words:
        if len(word) + len(currString) > maxSize:
            newLines.append(currString)
            currString = ''
        currString = currString + ' ' + word
    if currString:
        newLines.append(currString)

    return string.join(newLines,'\n')

srj = splitResizeJoin

def test():
    x = ZCVSMixin()
    x._v_debug = 1
    x.fsroot = '/tmp/zopeObjs/testCVS'
    x.fsrepos = ''

    a, theDict, b = x.cvsStatusList()

    print a, b
    for k in theDict.keys():
        print `k`
        for j in theDict[k].keys():
            print '     ',j,':', `theDict[k][j]`


if __name__=="__main__":
    test()
